------------------------------------------------------------------------
-- Event:        Delphi & Dintorni - July 09, 2020                     -
-- Session:      SQL Server Execution Plans: From Zero to Hero         -
--               https://bit.ly/3e9vLAB                                -
-- Demo:         Logical and physical operators (blue icons)           -
-- Author:       Sergio Govoni                                         -
-- Notes:        --                                                    -
------------------------------------------------------------------------

USE [AdventureWorks2017];
GO


-- Gli operatori logici e fisici descrivono il modo in cui
-- viene eseguita la query

-- Gli operatori logici derivano dai calcoli del query optimizer
-- per quello che potrebbe accadere in fase di esecuzione della query,
-- descrivono l'operazione algebrica relazionale

-- Gli operatori fisici rappresentano quello che  accaduto durante
-- l'esecuzione della query, descrivono l'algoritmo di implementazione
-- fisica, ad esempio Clustered Index Scan



/*** Clustered Index Scan ***/

SELECT
  *
FROM
  Person.Person;
GO

-- Tutta la struttura B-Tree dell'indice viene letta row-by-row,
-- l'optimizer valuta pi conveniente (= veloce) eseguire una
-- scansione rispetto all'utilizzo della chiave fornita dall'indice stesso

-- Clustered Index *Scan* ==> stiamo chiedendo a SQL Server
-- pi righe di quelle necessarie ?

-- While...End vs WHERE

-- La restituzione di righe non necessarie  uno spreco di risorse!!





/*** Clustered Index Seek ***/

SELECT
  P.*
FROM
  Person.Person AS P
WHERE
  (P.BusinessEntityID = 20);
GO

-- Se abbiamo la necessit di estrarre il contatto con ID = 20
-- applichiamo la clausola WHERE e rendiamo molto pi efficiente
-- la query precedente

-- I valori della chiave sono usati per identificare velocemente le righe





/*** Non-clustered Index Seek ***/

SELECT
  P.BusinessEntityID
FROM
  Person.Person AS P
WHERE
  (P.LastName like 'paul%');
GO

-- Uso dell'indice non-cluster per determinare le righe

-- In funzione della query e degli indici usati, l'optimizer
-- trover tutti i dati direttamente nell'indice non-cluster
-- oppure dovr eseguire un Key LookUp ==> I/O aggiuntive

-- Vediamo un Key LookUp...





/*** Key LookUp ***/

SELECT
  P.BusinessEntityID
  ,P.FirstName
  ,P.LastName
  ,P.EmailPromotion
  ,P.Title
FROM
  Person.Person AS P
WHERE
  (P.LastName like 'paul%');
GO

-- La prima operazione  un Index Seek

-- L'indice [IX_Person_LastName_FirstName_MiddleName] non "copre"
-- completamente la query, il QO  costretto a leggere l'indice cluster
-- (EmailPromotion e Title) ==> operazioni di I/O aggiuntive

-- Key Lookup  sempre accompagnato dall'operatore Nested Loops,
-- join che combina i risultati dei task precedenti

-- Le performance della query migliorano applicando un indice di "copertura"




-- Table Joins
SELECT
  emp.JobTitle
  ,a.City
  ,c.FirstName + ' ' + c.LastName AS Name
FROM
  HumanResources.Employee AS emp
JOIN
  Person.BusinessEntityAddress AS emp_adr ON emp_adr.BusinessEntityID=emp.BusinessEntityID
JOIN
  Person.Address AS a ON emp_adr.AddressID=a.AddressID
JOIN
  Person.Person AS c ON c.BusinessEntityID=emp.BusinessEntityID;
GO



/*** Hash Match (Join) ***/

-- Le operazioni di scan vengono combinate del Hash Match (Join)

-- Hashing and Hash Table ?

-- Una funzione Hash converte i dati in una forma simbolica,
-- ricerca facile e veloce, funzione reversibile

-- Una tabella di hash  una struttura dati che divide gli 
-- elementi (hashing) in categorie omogenee (= di pari dimensioni)

-- Hash Match (Join)  utilizzato quando SQL Server unisce due tabelle
-- attraverso una operazione di hashing sulle righe

-- E' molto efficiente quando una delle tabelle ha un numero di
-- righe notevolmente inferiore rispetto all'altra

-- La presenza di Hash Match pu indicare: 
-- 1. Un indice mancante o non utilizzabile
-- 2. Clausola WHERE mancante
-- 3. Clausola WHERE con CAST o CONVERT che disabilitano l'utilizzo di
--    eventuali indici

-- Una considerazione sul numero stimato di righe e numero effettivo di
-- righe...


/* Clustered Index Seek */

-- L'operazione pu costosa nel piano di esecuzione..

/* Nested Loops Join */

/* Compute Scalar */

-- Rappresenta semplicemente una operazione scalare, tipicamente un
-- calcolo o come nel nostro caso, la concatenazione della colonna
-- FirstName con la colonna LastName; Costo <> zero, ma trascurabile





/*** Merge Join ***/

SELECT
  c.CustomerID
FROM
  Sales.SalesOrderDetail AS sod
JOIN
  Sales.SalesOrderHeader AS soh ON sod.SalesOrderID=soh.SalesOrderID
JOIN
  Sales.Customer AS c ON c.CustomerID=soh.CustomerID;
GO

-- Scan per la mancanza della WHERE

-- Merge Join tra Customer e SalesOrderHeader

-- Merge Join  utilizzato quando le colonne in join sono pre-ordinate,
-- estremamente efficiente in questa situazione

-- L'optimizer pu scegliere di eseguire un ordinamento prima di applicare
-- un Merge Join, ma...

-- Applica Hash Join anche se meno efficiente...





-- GROUP BY and ORDER BY clause

/*** Sort ***/

SELECT
  s.*
FROM
  sales.salesorderheader AS s
ORDER BY
  s.orderdate;
GO

-- Costo dell Sort Task in relazione al costo totale query...

-- Se il Sort Task costa pi del 50% del costo totale
-- della query ==> ottimizzazione... manca la clausola WHERE?

-- E' davvero necessario ordinare il result-set?
-- Se non  necessario, evitiamo di applicare l'ORDER BY

SELECT
  s.*
FROM
  sales.salesorderheader AS s
ORDER BY
  s.salesorderid;
GO


-- Cambiando la colonna nell'ORDER BY non compare pi il Sort :) Perch?

-- Se dobbiamo ordinare grandi quantit di dati, verificare l'evento
-- "Sort Warning Event" (= Sort on tempdb), I/O...

-- Molti Sort Warning ==> aumento di RAM oppure migliorare l'accesso al tempdb!!



/*** Hash Match (Aggregate) ***/

SELECT
  a.PostalCode
  ,COUNT(a.PostalCode) AS PostalCount
FROM
  Person.Address a
GROUP BY
  a.PostalCode;
GO


-- No clausola WHERE ==> l'optimizer esegue la scansione

-- L'output viene passato a Hash Match Aggregate (<> dal join),
-- creazione di una tabella hash in memoria

-- L'aggregazione pu essere onerosa, l'unica ottimizzazione
--  verificare la presenza o meno della clausola WHERE





/*** Filter ***/

SELECT
  a.PostalCode
  ,COUNT(a.PostalCode) AS PostalCount
FROM
  Person.Address a
GROUP BY
  a.PostalCode
HAVING
  COUNT(a.PostalCode) > 210
GO


-- Clausola HAVING (per filtrare l'aggregazione), Filter nel piano
-- Filter  applicato SOLO dopo aver eseguito l'aggregazione!
-- Non miglioramenti di performance solo con l'utilizzo di HAVING, senza WHERE!
